-- Test routines for the connctrl package
--
-- The following functions test various aspects of the connctrl driver
-- using the connctrl package of helper functions.
--


local cc = require 'connctrl'

local function wait_for_controller(devname)
    local bar_map = cc.get_controller_map(devname)

    for i=1,100000 do if cc.is_controller_halted(devname) then break end end
    if not cc.is_controller_halted(devname) then
        print("\r\nFailed, breakpoint not reached")
        cc.dump_trace_buffer(devname)
        cc.controller_reset(devname)
        os.exit(1)
    end

    if bar_map:r32(4) == 0 then return end

    -- Exception occurred
    print("\r\nFailed, exception taken")
    cc.dump_exception_info(devname)

    cc.controller_reset(devname)
    os.exit(1)
end

local function test_controller_dpram(devname)
    local bar_map = cc.get_controller_map(devname)

    io.write("Controller Dual Port RAM Test ("..devname..") - ")

    cc.controller_reset(devname)

    -- Test byte read/write operations
    for i=0,15 do bar_map:w8(i, i) end
    for i=0,15 do
        if i ~= bar_map:r8(i) then
            print("Failed byte read/write test");
            os.exit(1)
        end
    end

    -- Test basic half-word read/write operations
    for i=0,14,2 do bar_map:w16(i, i * 0x102) end
    for i=0,14,2 do
        if (i * 0x102) ~= bar_map:r16(i) then
            print("Failed half-word read/write test");
            os.exit(1)
        end
    end

    -- Test basic word read/write operations
    for i=0,12,4 do bar_map:w32(i, i * 0x1020304) end
    for i=0,12,4 do
        if (i * 0x1020304) ~= bar_map:r32(i) then
            print("Failed word read/write test");
            os.exit(1)
        end
    end

    -- Test unaligned half-word read/write operations
    for i=0,14 do
        bar_map:w16(i, i * 0x102)
        if (i * 0x102) ~= bar_map:r16(i) then
            print("Failed unaligned half-word read/write test");
            os.exit(1)
        end
    end

    -- Test unaligned word read/write operations
    for i=0,12 do
        bar_map:w32(i, i * 0x1020304)
        if (i * 0x1020304) ~= bar_map:r32(i) then
            print("Failed unaligned word read/write test");
            os.exit(1)
        end
    end

    print("Passed")
end

local function test_controller_execution(devname)
    local bar_map = cc.get_controller_map(devname)

    io.write("Controller Execution Test ("..devname..") - ")

    cc.controller_reset(devname)
    cc.load_microcode(devname, "/usr/share/fpga-firmware/test1.bin")

    cc.release_controller_reset(devname)

    wait_for_controller(devname)

    cc.controller_reset(devname)

    local offset = bar_map:r32(0x14)

    for i=1,16 do
        if i ~= bar_map:r8(offset + (i - 1)) then
            print("Failed, execution results incorrect")
            os.exit(1)
        end
    end

    print("Passed")
end

local function dump_memory(map, offset, copy_size)
    local adjusted_offset = offset - (offset % 4) - 8
    local length = 16 + copy_size + (offset % 4) + 3
    for i = adjusted_offset,adjusted_offset+length,4 do
        io.write(string.format("%04X:%08X ", i, map:r32(i)))
        for j = i,i+3 do
            io.write(string.format("%02X ", map:r8(j)))
        end
        io.write("\r\n")
    end
end

local function error_dump(dma_map, offset_in_dma_page,
                          bar_map, offset_in_bar, chunk)
    print("Offset in bar is "..offset_in_bar)
    print("BAR memory")
    dump_memory(bar_map, offset_in_bar, #chunk)
    print("DMA memory")
    dump_memory(dma_map, offset_in_dma_page, #chunk)
end

local function test_dma_fpga_to_host(devname, requested_offset)
    local dma_map = cc.get_dma_mapping(devname, 65536)
    local bar_map = cc.get_controller_map(devname)
    local offset = requested_offset or 28
    local pass = 0

    io.write("DMA FPGA to Host Test ("..devname..") - ")

    -- Load valid code and let processor run until halted
    cc.controller_reset(devname)
    cc.load_microcode(devname, "/usr/share/fpga-firmware/test1.bin")
    cc.release_controller_reset(devname)
    wait_for_controller(devname)

    local controller_memory_size = cc.get_controller_memory_size(devname)
    if controller_memory_size > 4096 then controller_memory_size = 4096 end

    local chunk = "a"
    local filler = string.char(0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF)

    while #chunk < 27 do
        for offset_in_bar = 16,controller_memory_size-54,7 do
            local offset_in_dma_page = offset_in_bar + offset

            pass = pass + 1

            -- pad the source with marker bytes front and back
            bar_map:ws(offset_in_bar - #filler, filler)
            bar_map:ws(offset_in_bar, chunk)
            bar_map:ws(offset_in_bar + #chunk, filler)

            local zeroed = string.rep(string.char(0), #chunk + 2 * #filler)

            -- zero the destination including bytes front and back
            dma_map:ws(offset_in_dma_page - #filler, zeroed)

            if not cc.dma_host_transaction(devname,
                                           offset_in_bar,
                                           dma_map,
                                           offset_in_dma_page,
                                           #chunk,
                                           true) then
                print("Failed, dma_host_transaction() returned false")
                os.exit(1)
            end

            local compare = dma_map:rs(offset_in_dma_page - #filler,
                                       #chunk + 2 * #filler)

            if compare:sub(1 + #filler, #filler + #chunk) ~= chunk then
                print("Failed, message not transferred, pass "..pass)
                error_dump(dma_map, offset_in_dma_page,
                           bar_map, offset_in_bar, chunk)
                os.exit(1)
            end

            if compare:sub(1, #filler) ~= zeroed:sub(1, #filler) or
               compare:sub(1 + #filler + #chunk) ~= zeroed:sub(1, #filler) then
                print("Failed, destination guard bytes disturbed, pass "..pass)
                error_dump(dma_map, offset_in_dma_page,
                           bar_map, offset_in_bar, chunk)
                os.exit(1)
            end
        end

        chunk = chunk..string.char(string.byte(chunk, #chunk) + 1)
    end

    print("Passed")

    cc.controller_reset(devname)
    cc.release_dma_mapping(dma_map)
end

local function test_dma_host_to_fpga(devname, requested_offset)
    local dma_map = cc.get_dma_mapping(devname, 65536)
    local bar_map = cc.get_controller_map(devname)
    local offset = requested_offset or 28
    local pass = 0

    io.write("DMA Host to FPGA Test ("..devname..") - ")

    -- Load valid code and let processor run until halted
    cc.controller_reset(devname)
    cc.load_microcode(devname, "/usr/share/fpga-firmware/test1.bin")
    cc.release_controller_reset(devname)
    wait_for_controller(devname)

    local controller_memory_size = cc.get_controller_memory_size(devname)
    if controller_memory_size > 4096 then controller_memory_size = 4096 end

    local chunk = "a"
    local filler = string.char(0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF)

    while #chunk < 27 do
        for offset_in_bar = 16,controller_memory_size-54,7 do
            local offset_in_dma_page = offset_in_bar + offset

            pass = pass + 1

            -- pad the source with marker bytes front and back
            dma_map:ws(offset_in_dma_page - #filler, filler)
            dma_map:ws(offset_in_dma_page, chunk)
            dma_map:ws(offset_in_dma_page + #chunk, filler)

            local zeroed = string.rep(string.char(0), #chunk + 2 * #filler)

            -- zero the destination including bytes front and back
            bar_map:ws(offset_in_bar - #filler, zeroed)

            if not cc.dma_host_transaction(devname,
                                           offset_in_bar,
                                           dma_map,
                                           offset_in_dma_page,
                                           #chunk,
                                           false) then
                print("Failed, dma_host_transaction() returned false")
                os.exit(1)
            end

            local compare = bar_map:rs(offset_in_bar - #filler,
                                       #chunk + 2 * #filler)

            if compare:sub(1 + #filler, #filler + #chunk) ~= chunk then
                print("Failed, message not transferred, pass "..pass)
                error_dump(dma_map, offset_in_dma_page,
                           bar_map, offset_in_bar, chunk)
                os.exit(1)
            end

            if compare:sub(1, #filler) ~= zeroed:sub(1, #filler) or
               compare:sub(1 + #filler + #chunk) ~= zeroed:sub(1, #filler) then
                print("Failed, destination guard bytes disturbed, pass "..pass)
                error_dump(dma_map, offset_in_dma_page,
                           bar_map, offset_in_bar, chunk)
                os.exit(1)
            end
        end

        chunk = chunk..string.char(string.byte(chunk, #chunk) + 1)
    end

    print("Passed")

    cc.controller_reset(devname)
    cc.release_dma_mapping(dma_map)
end

local function test_controller_dma(devname)
    local dma_map = cc.get_dma_mapping(devname, 4096)
    local bar_map = cc.get_controller_map(devname)

    io.write("Controller DMA Test ("..devname..") - ")

    cc.controller_reset(devname)
    cc.load_microcode(devname, "/usr/share/fpga-firmware/test2.bin")

    bar_map:w32(0x14, cc.get_dma_mapping_physical_address(dma_map))

    for length = 16,1024,23 do
        for offset = 0x10,0x100,7 do
            local dest = 2048 + offset
            dma_map:w32(0, offset)
            dma_map:w32(4, length)
            dma_map:w32(8, dest)

            local byte = 1
            for i=1,length do
                dma_map:w8(offset + i - 1, byte)
                byte = byte + 1
                if byte == 254 then byte = 1 end
            end

            cc.release_controller_reset(devname)

            wait_for_controller(devname)

            for i=0,length-1 do
                if dma_map:r8(dest + i) ~= dma_map:r8(offset + i) + 1 then
                    print("Failed, execution results incorrect")
                    print(i, dma_map:r8(dest + i), dma_map:r8(offset + i))
                    print(offset, length, dest)
                    cc.controller_reset(devname)
                    cc.release_dma_mapping(dma_map)
                    os.exit(1)
                end
            end

            cc.controller_reset(devname)
        end
    end

    print("Passed")

    cc.controller_reset(devname)
    cc.release_dma_mapping(dma_map)
end

local function test_controller_interrupt(devname)
    io.write("Controller Interrupt Test ("..devname..") - ")

    cc.controller_reset(devname)
    cc.load_microcode(devname, "/usr/share/fpga-firmware/test3.bin")
    cc.clear_interrupts(devname)

    if cc.test_for_interrupt(devname) then
        print("Failed, interrupt already pending")
        cc.controller_reset(devname)
        os.exit(1)
    end

    cc.release_controller_reset(devname)

    for i=1,10000 do if cc.is_controller_halted(devname) then break end end
    if not cc.is_controller_halted(devname) then
        print("Failed, breakpoint not reached")
        cc.controller_reset(devname)
        os.exit(1)
    end

    if not cc.test_for_interrupt(devname) then
        print("Failed, interrupt not pending")
        cc.controller_reset(devname)
        os.exit(1)
    end

    print("Passed")

    cc.controller_reset(devname)
end

local function test_wait_for_irq(devname)
    io.write("Controller ("..devname..") - Waiting for IRQ - ")

    cc.wait_for_interrupt(devname)

    print("Done, Interrupted Received")
    os.exit(0)
end

local function test_issue_irq(devname)
    cc.controller_reset(devname)
    cc.load_microcode(devname, "/usr/share/fpga-firmware/test3.bin")
    cc.release_controller_reset(devname)
    cc.controller_run(devname)

    for i=1,10000 do if cc.is_controller_halted(devname) then break end end
    if not cc.is_controller_halted(devname) then
        print("Failed, breakpoint not reached")
        cc.controller_reset(devname)
        os.exit(1)
    end

    cc.controller_reset(devname)
    os.exit(0)
end

if #arg < 1 then
    print("Usage connctrl_test <device name> [<loop count>|IRQ_PEND|IRQ_ISSUE]")
    os.exit(1)
end

local loop_count = 1
if #arg > 1 then
    if arg[2] == "IRQ_PEND" then test_wait_for_irq(arg[1]) end
    if arg[2] == "IRQ_ISSUE" then test_issue_irq(arg[1]) end
    loop_count = tonumber(arg[2])
end

for i=1,loop_count do
    test_controller_dpram(arg[1])
    test_controller_execution(arg[1])
    test_dma_fpga_to_host(arg[1])
    test_dma_host_to_fpga(arg[1])
    test_controller_dma(arg[1])
    test_controller_interrupt(arg[1])
end

