require 'strict'

-- profile_generator_channel.lua - Implements the Time computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local rational = require 'rational'
local __ = require '__'
local log = require 'log'
local notify = require 'sys_notify'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list


local mt = {}

function mt:dump()
    dbg('tot_count=%lld, scale_factor=%lld',
        self.tot_count, self.scale_factor)
end
function getValue( file, x, scale, offset, drift, decay, size)
    scale = (decay^(x/size)) * scale
    offset = drift * (x/size) + offset 
    x = x%size
    return scale * (file[x + 1] or 0) + offset
end
-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer - NOT USED by this module!
-- Return values:  true (does nothing, for this module)
function mt:setup(sie_writer)
    return true
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    if not Pipe_has_frames(self.in_pipe) then
        return
    end

    -- Process each frame in input pipe.  There is only one input pipe.
    local of        -- will reference created output frame
    local out_vals  -- will reference "pointer" to output frame's data
    local _         -- we don't need 'in_vals' "pointer"
    local f = Pipe_get_frame(self.in_pipe)
    while f ~= nil do
        -- create frame for output
        of, _, out_vals = utils.create_output_frame(f.maxsize, f,
            self.tot_count, self.out_frame_sample_size)
        -- set output values
        local count = f.count
        for i = 0, count - 1 do
            if self.tot_count + i < self.file_lines * self.passes then
--            out_vals[i] = --self.scale * self.decay * 
                self.last_value = getValue(self.profile_file,(self.tot_count + i), self.scale, self.offset, self.drift, self.decay, self.file_lines)
                out_vals[i] = self.last_value
            else out_vals[i] = self.last_value
            end    
        end
        of.count = count
        self.tot_count = self.tot_count + of.count
        -- put output frame in output pipe(s)
        if not utils.put_frame(of, self.pipe_list) then
            Frame_delete(of)
        end
        -- delete input frame, get next
        Frame_delete(f)
        f = Pipe_get_frame(self.in_pipe)
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local profile_generator_channel = setmetatable({}, {__index = mt})
    profile_generator_channel.sie_collect_idx = first_sie_index
    utils.set_common_meta(profile_generator_channel, stanza)
    -- validate needed metadata, and create input/output pipe
    -- structures
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or
        stanza.summary.input_channels or {}
    assert(#input_channels == 1, "Time computed channel only" ..
        " accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    profile_generator_channel.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(profile_generator_channel.in_pipe and profile_generator_channel.in_pipe ~= nil,
        "failed to create input pipe")
    profile_generator_channel.pipe_list = Pipe_new_pipe_list()
    assert(profile_generator_channel.pipe_list and profile_generator_channel.pipe_list ~= nil,
        "failed to create pipe list")

    -- get remaining needed metadata
    local sample_rate =
        rational:new(stanza.sample_rate):tonumbers() or
        rational:new(stanza.summary.sample_rate):tonumbers()
    assert(sample_rate and sample_rate > 0,
        string.format("Invalid input sample rate: %s",
        tostring(sample_rate)))
    assert(stanza.data_type, "Missing data type")
    profile_generator_channel.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(profile_generator_channel.out_frame_sample_size and
           profile_generator_channel.out_frame_sample_size > 0,
        string.format("Invalid frame sample size:  %s",
        tostring(profile_generator_channel.out_frame_sample_size)))

    profile_generator_channel.scale_factor = stanza.timescale or (1 / sample_rate)
    profile_generator_channel.scale_factor = tonumber(profile_generator_channel.scale_factor)
    profile_generator_channel.profile_file = {}
    assert(tostring(stanza.profile_file), "Invalid filename")
    io.input("/root/" .. stanza.profile_file)
    profile_generator_channel.file_lines = 0
    local line_nbr = 0
    local valid_str
    for line in io.lines() do
        line_nbr = line_nbr + 1
        valid_str = string.match(line, "^[%+%-]?%d+%.?%d*[eE]?[%-%+]?%d*%c%c?$")
        if valid_str then
            -- Prevent a 0x1A character from being "read in".
            -- Some files may have a 0x1A as the last character
            -- in the file to make them compatible with some old
            -- Windows/DOS programs (possibly TCE).
            if string.byte(line, #line) == 26 then
                line = string.sub(line, 1, #line -1) .. " "
            elseif string.byte(line, #line - 1) == 26 then
                line = string.sub(line, 1, #line - 2) .. " "
            end
            table.insert(profile_generator_channel.profile_file, tonumber(line))
            profile_generator_channel.file_lines =
                profile_generator_channel.file_lines + 1
        else
            local sn = notify:new(__'Test Error')
            local msg = string.format(__'Profile generator file ' ..
                '"%s": text string at line number %s cannot be ' ..
                'converted to a number',
                tostring(stanza.profile_file), tostring(line_nbr))
            sn:error(msg)
            log:error(msg)
        end
    end

    assert(tonumber(stanza.scale), "Invalid scale")
    profile_generator_channel.scale = stanza.scale
    
    assert(tonumber(stanza.offset), "Invalid offset")
    profile_generator_channel.offset = stanza.offset

    assert(tonumber(stanza.drift), "Invalid drift")
    profile_generator_channel.drift = stanza.drift

    assert(tonumber(stanza.decay), "Invalid decay")
    profile_generator_channel.decay = stanza.decay

    assert(tonumber(stanza.passes), "Invalid Cycles")
    profile_generator_channel.passes = stanza.passes

    -- used for frames' time stamps (just a sample count for
    -- this module)
    profile_generator_channel.tot_count = 0

    return profile_generator_channel, 1
end

return m
